﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace BézierZeichenprogramm
{
    public partial class MainForm : Form
    {
        private List<BezierPath> paths = new List<BezierPath>();

        private int selectedCurve = -1;
        private int selectedPoint = -1;
        private Point prevPosition;

        public MainForm()
        {
            InitializeComponent();

            PointF[] p = new PointF[] { new PointF(200, 150), new PointF(230, 150), new PointF(250, 170), new PointF(250, 200), new PointF(250, 230), new PointF(230, 250), new PointF(200, 250) };
            BezierPath bp = new BezierPath(p);
            paths.Add(bp);

            p = new PointF[] { new PointF(210, 225), new PointF(230, 250), new PointF(350, 270), new PointF(150, 300), new PointF(250, 130), new PointF(130, 250) };
            bp = new BezierPath(p);
            paths.Add(bp);

            p = new PointF[] { new PointF(43, 224), new PointF(23, 124), new PointF(424, 625), new PointF(314, 213), new PointF(212, 412), new PointF(34, 23), new PointF(212, 412) };
            bp = new BezierPath(p);
            paths.Add(bp);

            p = new PointF[] { new PointF(323, 234), new PointF(23, 224), new PointF(44, 65), new PointF(34, 93), new PointF(212, 412) };
            bp = new BezierPath(p);
            paths.Add(bp);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
            e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;

            foreach (BezierPath p in paths)
            {
                p.Draw(e.Graphics);
            }

            if (selectedCurve >= 0)
            {
                paths[selectedCurve].DrawSelection(e.Graphics);
            }
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (CurveButton.Checked)
            {
                Bitmap bmp = new Bitmap(ClientSize.Width, ClientSize.Height);
                Graphics g = Graphics.FromImage(bmp);
                g.Clear(Color.Black);

                for (int i = 0; i < paths.Count; i++)
                {
                    BezierPath p = paths[i];
                    p.Draw(g, i);
                }

                Color c = bmp.GetPixel(e.X, e.Y);
                bmp.Dispose();

                int j = c.ToArgb() & 0x00FFFFFF;

                selectedCurve = j - 1;
            }
            else if(PointButton.Checked && selectedCurve >= 0)
            {
                selectedPoint = paths[selectedCurve].FindPoint(e.Location);
            }

            prevPosition = e.Location;

            Invalidate();
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (e.Button == MouseButtons.Left)
            {
                int dx = e.X - prevPosition.X;
                int dy = e.Y - prevPosition.Y;

                if (CurveButton.Checked && selectedCurve >= 0)
                {
                    paths[selectedCurve].Translate(dx, dy);
                }
                else if (PointButton.Checked && selectedCurve >= 0 && selectedPoint >= 0)
                {
                    paths[selectedCurve].TranslatePoint(selectedPoint, dx, dy);
                }

                prevPosition = e.Location;

                Invalidate();
            }
        }

        private void CurveButton_Click(object sender, EventArgs e)
        {
            CurveButton.Checked = true;
            PointButton.Checked = false;
        }

        private void PointButton_Click(object sender, EventArgs e)
        {
            CurveButton.Checked = false;
            PointButton.Checked = true;
        }
    }

    internal class BezierPath
    {
        private List<PointF> points = new List<PointF>();

        private Pen pen = Pens.Black;
        private Brush brush = Brushes.Red;
        static private Random random = new Random();

        internal BezierPath(PointF[] Points)
        {
            this.points.AddRange(Points);
            int r = random.Next(190); // not too bright
            int g = random.Next(190);
            int b = random.Next(190);
            Color c = Color.FromArgb(r, g, b);
            pen = new Pen(c, 2.0f);
            brush = new SolidBrush(c);
        }

        internal void Draw(Graphics g)
        {
            if (points.Count >= 4)
            {
                int numValid = 3 * ((points.Count - 4) / 3) + 4;
                PointF[] validPoints = points.GetRange(0, numValid).ToArray();
                g.DrawBeziers(pen, validPoints);
            }
        }

        internal void Draw(Graphics g, int i)
        {
            if (i < 0 || i >= 0x00FFFFFF)
            {
                throw new Exception("too many curves");
            }

            if (points.Count >= 4)
            {
                Color c = Color.FromArgb(i+1+(0xFF<<24));
                int numValid = 3 * ((points.Count - 4) / 3) + 4;
                PointF[] validPoints = points.GetRange(0, numValid).ToArray();
                g.DrawBeziers(new Pen(c, 8.0f), validPoints);
            }
        }

        internal void DrawSelection(Graphics g)
        {
            foreach (PointF p in points)
            {
                g.FillEllipse(brush, p.X - 4, p.Y - 4, 8, 8);
            }
        }

        internal void Translate(int dx, int dy)
        {
            for(int i = 0; i < points.Count; i++)
            {
                PointF p = points[i];
                points[i] = new PointF(p.X + dx, p.Y + dy);
            }
        }

        internal int FindPoint(PointF p)
        {
            for (int i = 0; i < points.Count; i++)
            {
                if (new RectangleF(new PointF(points[i].X - 3, points[i].Y - 3), new Size(6, 6)).Contains(p))
                {
                    return i;
                }
            }
            return -1;
        }

        internal void TranslatePoint(int selectedPoint, int dx, int dy)
        {
            float px0 = points[selectedPoint].X;
            float py0 = points[selectedPoint].Y;
            float px1 = px0 + dx;
            float py1 = py0 + dy;
            points[selectedPoint] = new PointF(px1, py1);

            bool firstHandle = (selectedPoint % 3 == 1);
            bool secondHandle = (selectedPoint % 3 == 2);
            if (firstHandle && selectedPoint > 3 || secondHandle && points.Count > selectedPoint + 2) // Es gibt einen Partnerpunkt.
            {
                int center = selectedPoint + 1;
                if (firstHandle)
                {
                    center = selectedPoint - 1;
                }

                double phi0 = Math.Atan2(py0 - points[center].Y, px0 - points[center].X);
                double phi1 = Math.Atan2(py1 - points[center].Y, px1 - points[center].X);
                double phi = phi1 - phi0;
                //System.Diagnostics.Debug.WriteLine(phi0.ToString() + " " + phi1.ToString());

                int partner = selectedPoint + 2;
                if(firstHandle)
                {
                    partner = selectedPoint - 2;
                }

                float qx0 = points[partner].X - points[center].X;
                float qy0 = points[partner].Y - points[center].Y;

                double s = Math.Sin(phi);
                double c = Math.Cos(phi);

                double qx1 = points[center].X + c * qx0 - s * qy0;
                double qy1 = points[center].Y + s * qx0 + c * qy0;

                points[partner] = new PointF((float)qx1, (float)qy1);
            }
        }
    }
}